"""
Utilities to read the input data and format them in a way to be read by
volume_matching.

@title: io
@author: Valentin Louf <valentin.louf@bom.gov.au>
@institutions: Monash University and the Australian Bureau of Meteorology
@creation: 17/02/2020
@date: 12/12/2025

.. autosummary::
    :toctree: generated/

    NoPrecipitationError
    _read_radar
    check_precip_in_domain
    data_load_and_checks
    get_ground_radar_attributes
    get_gpm_orbit
    read_GPM
    read_radars
"""

import re
import copy
import datetime
import warnings

from typing import Tuple, List, Union
from collections import OrderedDict

import h5py
import pyodim
import pyproj
import numpy as np
import pandas as pd
import xarray as xr

from . import correct
from . import default


class NoPrecipitationError(Exception):
    pass


def check_precip_in_domain(
    gpmset: xr.Dataset, grlon: float, grlat: float, rmax: float = 150e3, rmin: float = 20e3
) -> Tuple[pd.Timestamp, int]:
    """
    Check if there's GPM precipitation in the radar domain.

    Parameters:
    ===========
    gpmset: xarray
        Dataset containing the GPM data
    grlon: float
        Ground radar longitude
    grlat: float
        Ground radar latitude
    rmax: float
        Ground radar maximum range
    rmin: float
        Ground radar minimum range.

    Returns:
    ========
    gpmtime0: datetime
        Time of the closest measurement from ground radar site.
    nprof: int
        Number of GPM precipitation profiles in ground radar domain.
    """
    georef = pyproj.Proj(f"+proj=aeqd +lon_0={grlon} +lat_0={grlat} +ellps=WGS84")
    gpmlat = gpmset.Latitude.values
    gpmlon = gpmset.Longitude.values

    xgpm, ygpm = georef(gpmlon, gpmlat)
    rproj_gpm = (xgpm**2 + ygpm**2) ** 0.5

    gr_domain = (rproj_gpm <= rmax) & (rproj_gpm >= rmin)
    if gr_domain.sum() < 10:
        raise NoPrecipitationError("GPM swath does not go through the radar domain.")

    nprof = np.sum(gpmset.flagPrecip.values[gr_domain])
    if nprof < 10:
        raise NoPrecipitationError("No precipitation measured by GPM inside radar domain.")

    newset = gpmset.merge({"range_from_gr": (("nscan", "nray"), rproj_gpm)})
    gpmtime0 = newset.nscan.where(newset.range_from_gr == newset.range_from_gr.min()).values.astype("datetime64[s]")
    gpmtime0 = gpmtime0[~np.isnat(gpmtime0)][0]
    gpmtime = pd.Timestamp(gpmtime0)
    del newset
    return gpmtime, nprof


def data_load_and_checks(
    gpmfile: str,
    grfile: str,
    refl_name: Union[str, None] = None,
    correct_attenuation: bool = True,
    radar_band: str = "C",
    kdp_name: Union[str, None] = "KDP",
    phase_aware_dfr: bool = True,
) -> Tuple[xr.Dataset, List[xr.Dataset]]:
    """
    Load GPM and Ground radar files and perform some initial checks:
    domains intersect, precipitation, time difference.

    Parameters:
    -----------
    gpmfile: str
        GPM data file.
    grfile: str
        Ground radar input file.
    grfile2: str
        (Optional) Second ground radar input file to compute grid displacement
        and advection.
    refl_name: str
        Name of the reflectivity field in the ground radar data.
    correct_attenuation: bool
        Should we correct for C- or X-band ground radar attenuation.
    radar_band: str
        Ground radar frequency band for reflectivity conversion. S, C, and X
        supported.
    kdp_name: Optional[str]
        Name of the KDP field in the ground radar data for attenuation correction.
    phase_aware_dfr: bool
        Use phase-aware DFR conversion that accounts for ice, melting layer, and
        liquid precipitation phases using GPM bright band height. Default is True.
        Set to False to use the simple polynomial DFR relationship.

    Returns:
    --------
    gpmset: xarray.Dataset
        Dataset containing the input datas.
    radar: pyart.core.Radar
        Pyart radar dataset.
    """
    if refl_name is None:
        raise ValueError("Reflectivity field name not given.")

    gpmset = read_GPM(gpmfile)
    grlon, grlat, gralt, rmin, rmax = get_ground_radar_attributes(grfile)

    # Reproject satellite coordinates onto ground radar
    georef = pyproj.Proj(f"+proj=aeqd +lon_0={grlon} +lat_0={grlat} +ellps=WGS84")
    gpmlat = gpmset.Latitude.values
    gpmlon = gpmset.Longitude.values

    xgpm, ygpm = georef(gpmlon, gpmlat)
    rproj_gpm = (xgpm**2 + ygpm**2) ** 0.5

    gr_domain = (rproj_gpm <= rmax) & (rproj_gpm >= rmin)
    if gr_domain.sum() < 10:
        info = f"The closest satellite measurement is {np.min(rproj_gpm / 1e3):0.4} km away from ground radar."
        if gr_domain.sum() == 0:
            raise NoPrecipitationError("GPM swath does not go through the radar domain. " + info)
        else:
            raise NoPrecipitationError("Not enough GPM precipitation inside ground radar domain. " + info)

    nprof = np.sum(gpmset.flagPrecip.values[gr_domain])
    if nprof < 10:
        raise NoPrecipitationError("No precipitation measured by GPM inside radar domain.")

    # Parallax correction
    sr_xp, sr_yp, z_sr = correct.correct_parallax(xgpm, ygpm, gpmset)

    # Compute the elevation of the satellite bins with respect to the ground radar.
    gr_gaussian_radius = correct.compute_gaussian_curvature(grlat)
    gamma = np.sqrt(sr_xp**2 + sr_yp**2) / gr_gaussian_radius
    elev_sr_grref = np.rad2deg(
        np.arctan2(np.cos(gamma) - (gr_gaussian_radius + gralt) / (gr_gaussian_radius + z_sr), np.sin(gamma))
    )

    # Convert reflectivity band correction
    if phase_aware_dfr:
        # Use phase-aware DFR that accounts for ice, melting layer, and liquid phases
        reflgpm_grband = correct.convert_gpmrefl_grband_phase_aware(
            refl_gpm=gpmset.zFactorCorrected.values,
            height_gpm=z_sr,
            height_bb=gpmset.heightBB.values,
            radar_band=radar_band,
        )
    else:
        # Use simple polynomial DFR relationship (original method)
        reflgpm_grband = correct.convert_gpmrefl_grband_dfr(gpmset.zFactorCorrected.values, radar_band=radar_band)

    gpmset = gpmset.merge(
        {
            "precip_in_gr_domain": (("nscan", "nray"), gpmset.flagPrecip.values & gr_domain),
            "range_from_gr": (("nscan", "nray"), rproj_gpm),
            "elev_from_gr": (("nscan", "nray", "nbin"), elev_sr_grref),
            "x": (("nscan", "nray", "nbin"), sr_xp),
            "y": (("nscan", "nray", "nbin"), sr_yp),
            "z": (("nscan", "nray", "nbin"), z_sr),
            "reflectivity_grband": (("nscan", "nray", "nbin"), reflgpm_grband),
        }
    )

    # Get time of the overpass (closest point from ground radar).
    gpmtime0 = gpmset.nscan.where(gpmset.range_from_gr == gpmset.range_from_gr.min()).values.astype("datetime64[s]")
    gpmtime0 = gpmtime0[~np.isnat(gpmtime0)][0]
    gpmset = gpmset.merge({"overpass_time": (gpmtime0)})

    # Attributes
    metadata = default.gpmset_metadata()
    for k, v in metadata.items():
        for sk, sv in v.items():
            try:
                gpmset[k].attrs[sk] = sv
            except KeyError:
                continue
    gpmset.reflectivity_grband.attrs["description"] = f"Satellite reflectivity converted to {radar_band}-band."
    gpmset.attrs["nprof"] = nprof
    gpmset.attrs["earth_gaussian_radius"] = gr_gaussian_radius

    # Time to read the ground radar data.
    radar = read_radar(
        grfile, refl_name, radar_band=radar_band, correct_attenuation=correct_attenuation, kdp_name=kdp_name
    )

    return gpmset, radar


def get_gpm_orbit(gpmfile: str) -> int:
    """
    Parameters:
    -----------
    gpmfile: str
        GPM data file.

    Returns:
    --------
    orbit: int
        GPM Granule Number.
    """
    try:
        with h5py.File(gpmfile) as hid:
            grannb = [s for s in hid.attrs["FileHeader"].split() if b"GranuleNumber" in s][0].decode("utf-8")
            orbit = re.findall("[0-9]{3,}", grannb)[0]
    except Exception:
        return 0

    return int(orbit)


def get_ground_radar_attributes(grfile: str) -> Tuple[float, float, float, float, float]:
    """
    Read the ground radar attributes, latitude/longitude, altitude, range
    min/max.

    Parameter:
    ==========
    grfile: str
        Input ground radar file.

    Returns:
    ========
    grlon: float
        Radar longitude.
    grlat: float
        Radar latitude.
    gralt: float
        Radar altitude.
    rmin : float
        Radar minimum range (cone of silence.)
    rmax: float
        Radar maximum range.
    """
    nradar = pyodim.read_odim(grfile)
    radar = nradar[0].compute()
    rmin = radar.range.values.min()
    if rmin < 15e3:
        rmin = 15e3
    rmax = radar.range.values.max()
    grlon = radar.attrs["longitude"]
    grlat = radar.attrs["latitude"]
    gralt = radar.attrs["height"]

    return grlon, grlat, gralt, rmin, rmax


def read_GPM(infile: str, refl_min_thld: float = 0) -> xr.Dataset:
    """
    Read GPM data and organize them into a Dataset.

    Parameters:
    -----------
    gpmfile: str
        GPM data file.
    refl_min_thld: float
        Minimum threshold applied to GPM reflectivity.

    Returns:
    --------
    dset: xr.Dataset
        GPM dataset.
    """
    if refl_min_thld != 0:
        warnings.warn("Tests have shown that no threshold should be applied to GPM reflectivity!", UserWarning)
    data = dict()
    date = dict()
    with h5py.File(infile, "r") as hid:
        try:
            master_key = "NS"
            keys = hid[f"/{master_key}"].keys()
        except KeyError:
            master_key = "FS"
            keys = hid[f"/{master_key}"].keys()

        for k in keys:
            if k == "Latitude" or k == "Longitude":
                dims = tuple(hid[f"/{master_key}/{k}"].attrs["DimensionNames"].decode("UTF-8").split(","))
                fv = hid[f"/{master_key}/{k}"].attrs["_FillValue"]
                data[k] = (dims, np.ma.masked_equal(hid[f"/{master_key}/{k}"][:], fv))
            else:
                try:
                    subkeys = hid[f"/{master_key}/{k}"].keys()
                except Exception:
                    continue
                for sk in subkeys:
                    dims = tuple(hid[f"/{master_key}/{k}/{sk}"].attrs["DimensionNames"].decode("UTF-8").split(","))
                    fv = hid[f"/{master_key}/{k}/{sk}"].attrs["_FillValue"]

                    if sk in ["Year", "Month", "DayOfMonth", "Hour", "Minute", "Second", "MilliSecond"]:
                        date[sk] = np.ma.masked_equal(hid[f"/{master_key}/{k}/{sk}"][:], fv)
                    elif sk in ["DayOfYear", "SecondOfDay"]:
                        continue
                    elif sk == "typePrecip":
                        # Simplify precipitation type
                        data[sk] = (dims, hid[f"/{master_key}/{k}/{sk}"][:] / 10000000)
                    elif sk in ["zFactorCorrected", "zFactorFinal", "zFactorMeasured"]:
                        # Reverse direction along the beam.
                        gpm_refl = hid[f"/{master_key}/{k}/{sk}"][:][:, :, ::-1]
                        gpm_refl[gpm_refl < 0] = np.nan
                        data[sk] = (dims, np.ma.masked_invalid(np.ma.masked_less_equal(gpm_refl, refl_min_thld)))
                    elif sk == "flagPrecip":
                        data[sk] = (
                            dims,
                            np.ma.masked_invalid(hid[f"/{master_key}/{k}/{sk}"][:]).filled(0).astype(bool),
                        )
                    else:
                        data[sk] = (dims, np.ma.masked_equal(hid[f"/{master_key}/{k}/{sk}"][:], fv))

    try:
        _ = data["zFactorCorrected"]
    except KeyError:
        data["zFactorCorrected"] = copy.deepcopy(data["zFactorFinal"])

    # Create Quality indicator.
    quality = np.zeros(data["heightBB"][-1].shape, dtype=np.int32)
    quality[((data["qualityBB"][-1] == 0) | (data["qualityBB"][-1] == 1)) & (data["qualityTypePrecip"][-1] == 1)] = 1
    quality[(data["qualityBB"][-1] > 1) | (data["qualityTypePrecip"][-1] > 1)] = 2
    data["quality"] = (data["heightBB"][0], quality)

    # Generate dimensions.
    nray = np.linspace(-17.04, 17.04, 49)
    nbin = np.arange(0, 125 * 176, 125)

    R, A = np.meshgrid(nbin, nray)
    distance_from_sr = 407000 / np.cos(np.deg2rad(A)) - R  # called rt in IDL code.
    data["distance_from_sr"] = (("nray", "nbin"), distance_from_sr)

    try:
        # TRMM doesn't have a MilliSecond field.
        _ = date["MilliSecond"]
    except KeyError:
        date["MilliSecond"] = date["Second"]

    dtime = np.array(
        [
            datetime.datetime(*d)
            for d in zip(
                date["Year"],
                date["Month"],
                date["DayOfMonth"],
                date["Hour"],
                date["Minute"],
                date["Second"],
                date["MilliSecond"],
            )
        ],
        dtype="datetime64",
    )

    data["nscan"] = (("nscan"), dtime)
    data["nray"] = (("nray"), nray)
    data["nbin"] = (("nbin"), nbin)

    dset = xr.Dataset(OrderedDict(sorted(data.items())))

    dset.nray.attrs = {"units": "degree", "description": "Deviation from Nadir"}
    dset.nbin.attrs = {"units": "m", "description": "Downward from 0: TOA to Earth ellipsoid."}
    dset.attrs["altitude"] = 407000
    dset.attrs["altitude_units"] = "m"
    dset.attrs["altitude_description"] = "GPM orbit"
    dset.attrs["beamwidth"] = 0.71
    dset.attrs["beamwidth_units"] = "degree"
    dset.attrs["beamwidth_description"] = "GPM beamwidth"
    dset.attrs["dr"] = 125
    dset.attrs["dr_units"] = "m"
    dset.attrs["dr_description"] = "GPM gate spacing"
    dset.attrs["orbit"] = get_gpm_orbit(infile)

    return dset


def read_radar(
    grfile: str,
    refl_name: str,
    radar_band: str = "C",
    correct_attenuation: bool = False,
    kdp_name: Union[str, None] = "KDP",
) -> List[xr.Dataset]:
    """
    Read ground radar data. If 2 files provided, then it will compute the
    displacement between these two files and then correct for advection the
    ground radar data in relation to the time of GPM exact overpass.

    Parameters:
    ===========
    grfile: str
        Ground radar input file.
    refl_name: str
        Name of the reflectivity field in the ground radar data.
    radar_band: str
        Ground radar frequency band for reflectivity conversion. S, C, and X
        supported.
    correct_attenuation: bool
        Should we correct for C- or X-band ground radar attenuation.
    kdp_name: Optional[str]
        Name of the KDP field in the ground radar data for attenuation correction.

    Returns:
    ========
    radar: List[xr.Dataset]
        Radar dataset
    """
    nradar = pyodim.read_odim(grfile, lazy_load=False)
    if nradar[-1].elevation.max() > 80:
        nradar.pop(-1)

    method = None
    if correct_attenuation and radar_band in ["X", "C"]:  # Correct attenuation of X or C bands.
        for idx, radar in enumerate(nradar):
            zh = radar[refl_name].values.copy()
            dr = 1e-3 * (radar.range.values[1] - radar.range.values[0])  # in km
            if kdp_name in radar:
                kdp = radar[kdp_name].values.copy()
                kdp[np.isnan(kdp)] = 0
                zh_corrected = correct.attenuation_correction_zphi(zh, kdp, dr=dr, wavelength=radar_band)
                method = "ZPHI"
            else:
                zh_corrected = correct.attenuation_correction_gunn_east(zh, dr=dr, wavelength=radar_band)
                method = "Gunn-East"

            nradar[idx] = nradar[idx].merge({"ZH_ATTEN_CORR": (("azimuth", "range"), zh_corrected)})

    if method is not None:
        print(f"GR reflectivity corrected for attenuation using {method} in band {radar_band}")

    return nradar



"""
GPM volume matching package.

Volume matching of ground radar and GPM satellite data.
"""

from .gpmmatch import (
    NoRainError,
    generate_filename,
    get_radar_coordinates,
    get_gr_reflectivity,
    has_valid_data,
    volume_matching,
    vmatch_multi_pass,
    MIN_SAMPLE_POINTS,
    MIN_REFL_SAMPLES,
    MAX_OFFSET_THRESHOLD,
)

__author__ = "Valentin Louf"
__email__ = "valentin.louf@bom.gov.au"

__all__ = [
    "NoRainError",
    "generate_filename",
    "get_radar_coordinates",
    "get_gr_reflectivity",
    "has_valid_data",
    "volume_matching",
    "vmatch_multi_pass",
    "MIN_SAMPLE_POINTS",
    "MIN_REFL_SAMPLES",
    "MAX_OFFSET_THRESHOLD",
]



"""
Volume matching of ground radar and GPM satellite. Default naming and attribute.

@title: gpmmatch
@author: Valentin Louf <valentin.louf@bom.gov.au>
@institutions: Monash University and the Australian Bureau of Meteorology
@creation: 24/02/2020
@date: 29/05/2025

.. autosummary::
    :toctree: generated/

    gpmset_metadata
    get_metadata
"""

from typing import Dict, Any


def gpmset_metadata() -> Dict[str, Any]:
    """
    Return a bunch of metadata (description, units, long_name, etc.) for the
    GPM set metadata.
    This metadata is used to describe the GPM data columns that are
    volume-matched to the ground radar data.
    The metadata includes information about the GPM overpass time, parallax
    corrected coordinates, precipitation inside the ground radar scope,
    range and elevation from the ground radar, and reflectivity in the
    ground radar band.

    Returns:
    ========
    metadata: dict
    """
    metadata = {
        "overpass_time": {"description": "GPM overpass time at the closest from ground radar site"},
        "x": {
            "units": "m",
            "description": "x-axis parallax corrected coordinates in relation to ground radar.",
        },
        "y": {
            "units": "m",
            "description": "y-axis parallax corrected coordinates in relation to ground radar.",
        },
        "z": {
            "units": "m",
            "description": "z-axis parallax corrected coordinates in relation to ground radar.",
        },
        "precip_in_gr_domain": {
            "units": "1",
            "description": "Satellite data-columns with precipitation inside the ground radar scope.",
        },
        "range_from_gr": {
            "units": "m",
            "description": "Range from satellite bins in relation to ground radar",
        },
        "elev_from_gr": {
            "units": "degrees",
            "description": "Elevation from satellite bins in relation to ground radar",
        },
        "reflectivity_grband": {"units": "dBZ"},
    }
    return metadata


def get_metadata() -> Dict[str, Dict[str, str]]:
    """
    Return a bunch of metadata (description, units, long_name, etc.) for the
    output dataset.

    Returns:
    ========
    metadata: dict
    """
    metadata = {
        "fmin_gpm": {
            "units": "",
            "long_name": "fmin_gpm",
            "description": "FMIN ratio for GPM.",
        },
        "fmin_gr": {
            "units": "",
            "long_name": "fmin_gr",
            "description": "FMIN ratio for GR.",
        },
        "refl_gpm_raw": {
            "units": "dBZ",
            "long_name": "GPM_reflectivity",
            "description": "GPM reflectivity volume-matched to ground radar.",
        },
        "refl_gpm_grband": {
            "units": "dBZ",
            "long_name": "GPM_reflectivity_grband_stratiform",
            "description": "GPM reflectivity converted to ground radar frequency band.",
        },
        "refl_gr_raw": {
            "units": "dBZ",
            "long_name": "reflectivity",
            "description": "Ground radar reflectivity volume matched using a `normal` average.",
        },
        "refl_gr_weigthed": {
            "units": "dBZ",
            "long_name": "reflectivity",
            "description": "Ground radar reflectivity volume matched using a distance-weighted average.",
        },
        "std_refl_gpm": {
            "units": "dB",
            "long_name": "standard_deviation_reflectivity",
            "description": "GPM reflectivity standard deviation of the volume-matched sample.",
        },
        "std_refl_gr": {
            "units": "dB",
            "long_name": "standard_deviation_reflectivity",
            "description": "Ground radar reflectivity standard deviation of the volume-matched sample.",
        },
        "sample_gpm": {
            "units": "1",
            "long_name": "sample_size",
            "description": "Number of GPM bins used to compute the volume-matched pixels at a given points",
        },
        "reject_gpm": {
            "units": "1",
            "long_name": "rejected_sample_size",
            "description": "Number of GPM bins rejected to compute the volume-matched pixels at a given points",
        },
        "sample_gr": {
            "units": "1",
            "long_name": "sample_size",
            "description": "Number of ground-radar bins used to compute the volume-matched pixels at a given points",
        },
        "reject_gr": {
            "units": "1",
            "long_name": "rejected_sample_size",
            "description": "Number of ground-radar bins rejected to compute the volume-matched pixels at a given points",
        },
        "volume_match_gpm": {
            "units": "m^3",
            "long_name": "volume",
            "description": "Volume of the GPM sample for each match points.",
        },
        "volume_match_gr": {
            "units": "m^3",
            "long_name": "volume",
            "description": "Volume of the ground radar sample for each match points.",
        },
        "x": {
            "units": "m",
            "long_name": "projected_x_axis_coordinates",
            "projection": "Azimuthal Equidistant from ground radar.",
        },
        "y": {
            "units": "m",
            "long_name": "projected_y_axis_coordinates",
            "projection": "Azimuthal Equidistant from ground radar.",
        },
        "z": {
            "units": "m",
            "long_name": "projected_z_axis_coordinates",
            "projection": "Azimuthal Equidistant from ground radar.",
        },
        "r": {
            "units": "m",
            "long_name": "range",
            "description": "Range from ground radar.",
        },
        "timedelta": {
            "units": "ns",
            "long_name": "timedelta",
            "description": "Maximum time delta between ground radar and GPM volumes.",
        },
        "elevation_gr": {
            "units": "degrees",
            "long_name": "elevation",
            "description": "Ground radar reference elevation.",
        },
        "ntilt": {
            "units": "1",
            "long_name": "ground_radar_tilt_number",
            "description": "Number of ground radar tilts used for volume matching.",
        },
        "nprof": {
            "units": "1",
            "long_name": "gpm_profile_number",
            "description": "Number of GPM profiles (nrays x nscan) used for volume matching.",
        },
        "pir_gpm": {
            "units": "dB m-1",
            "long_name": "GPM_path_integrated_reflectivity",
            "description": "Path integrated GPM reflectivity volume-matched.",
        },
        "pir_gr": {
            "units": "dB m-1",
            "long_name": "GR_path_integrated_reflectivity",
            "description": "Path integrated GR reflectivity volume-matched.",
        },
    }
    return metadata



"""
Unit tests for gpmmatch module.

Tests volume matching of ground radar and GPM satellite data.
"""

import shutil
import pytest
import tempfile
from pathlib import Path
import numpy as np
import pandas as pd
import xarray as xr

# Import from the specific module file, not the package
from gpmmatch.gpmmatch import (
    generate_filename,
    get_radar_coordinates,
    get_gr_reflectivity,
    has_valid_data,
    volume_matching,
    vmatch_multi_pass,
)


@pytest.fixture
def sample_radar_dataset():
    """Create a minimal radar dataset for testing."""
    ntime = 10
    nrange = 50

    radar = xr.Dataset(
        {
            "reflectivity": (["time", "range"], np.random.uniform(10, 50, (ntime, nrange))),
            "corrected_reflectivity": (["time", "range"], np.random.uniform(10, 50, (ntime, nrange))),
        },
        coords={
            "time": pd.date_range("2018-11-20 05:36:30", periods=ntime, freq="10s"),
            "range": np.arange(1000, 1000 + nrange * 250, 250),
            "x": (["time", "range"], np.random.uniform(-50000, 50000, (ntime, nrange))),
            "y": (["time", "range"], np.random.uniform(-50000, 50000, (ntime, nrange))),
            "azimuth": (["time"], np.linspace(0, 360, ntime)),
        },
        attrs={
            "start_time": "2018-11-20T05:36:30Z",
            "end_time": "2018-11-20T05:38:00Z",
            "longitude": 145.0,
            "latitude": -37.8,
        }
    )

    radar["elevation"] = (["time"], np.ones(ntime) * 0.5)

    return radar


@pytest.fixture
def sample_gpm_dataset():
    """Create a minimal GPM dataset for testing."""
    nscan = 20
    nray = 49
    nbin = 176

    gpm = xr.Dataset(
        {
            "zFactorCorrected": (["nscan", "nray", "nbin"],
                                np.random.uniform(10, 45, (nscan, nray, nbin))),
            "reflectivity_grband": (["nscan", "nray", "nbin"],
                                   np.random.uniform(10, 45, (nscan, nray, nbin))),
            "precip_in_gr_domain": (["nscan", "nray"],
                                   np.random.choice([0, 1], size=(nscan, nray), p=[0.3, 0.7])),
            "elev_from_gr": (["nscan", "nray", "nbin"],
                           np.random.uniform(0.5, 2.5, (nscan, nray, nbin))),
            "flagPrecip": (["nscan", "nray"], np.random.choice([0, 1, 2], size=(nscan, nray))),
        },
        coords={
            "nscan": pd.date_range("2018-11-20 05:33:41", periods=nscan, freq="1s"),
            "nray": np.arange(nray),
            "x": (["nscan", "nray", "nbin"], np.random.uniform(-100000, 100000, (nscan, nray, nbin))),
            "y": (["nscan", "nray", "nbin"], np.random.uniform(-100000, 100000, (nscan, nray, nbin))),
            "z": (["nscan", "nray", "nbin"], np.random.uniform(1000, 20000, (nscan, nray, nbin))),
        },
        attrs={
            "orbit": 26860,
        }
    )

    gpm["dr"] = 125.0
    gpm["beamwidth"] = 0.71
    gpm["altitude"] = 407000.0
    gpm["earth_gaussian_radius"] = 6371000.0
    gpm["distance_from_sr"] = (["nbin"], np.arange(nbin) * 125.0)
    gpm["overpass_time"] = pd.Timestamp("2018-11-20 05:37:00")

    return gpm


@pytest.fixture
def temp_output_dir():
    """Create a temporary directory for outputs."""
    temp_dir = tempfile.mkdtemp()
    yield Path(temp_dir)
    shutil.rmtree(temp_dir)


# Test helper functions
class TestHelperFunctions:

    def test_generate_filename(self, sample_radar_dataset, sample_gpm_dataset):
        """Test filename generation."""
        filename = generate_filename(sample_radar_dataset, sample_gpm_dataset, "test_radar")

        assert "vmatch.gpm.orbit" in filename
        assert "0026860" in filename
        assert "test_radar" in filename
        assert ".nc" in filename
        assert "20181120" in filename

    def test_get_radar_coordinates_no_offset(self, sample_radar_dataset):
        """Test radar coordinate extraction without elevation offset."""
        nradar = [sample_radar_dataset]

        range_gr, elev_gr, xradar, yradar, time_radar = get_radar_coordinates(nradar)

        assert len(range_gr) == 50
        assert len(elev_gr) == 1
        assert elev_gr[0] == pytest.approx(0.5)
        assert len(xradar) == 1
        assert len(yradar) == 1
        assert len(time_radar) == 1

    def test_get_radar_coordinates_with_offset(self, sample_radar_dataset):
        """Test radar coordinate extraction with elevation offset."""
        nradar = [sample_radar_dataset]
        elevation_offset = 0.3

        range_gr, elev_gr, xradar, yradar, time_radar = get_radar_coordinates(
            nradar, elevation_offset=elevation_offset
        )

        assert elev_gr[0] == pytest.approx(0.8)

    def test_get_gr_reflectivity(self, sample_radar_dataset):
        """Test ground radar reflectivity extraction."""
        nradar = [sample_radar_dataset]

        refl, pir = get_gr_reflectivity(nradar, "corrected_reflectivity", gr_offset=0, gr_refl_threshold=10)

        assert len(refl) == 1
        assert len(pir) == 1
        assert isinstance(refl[0], np.ma.MaskedArray)
        assert refl[0].shape == (10, 50)

    def test_get_gr_reflectivity_with_offset(self, sample_radar_dataset):
        """Test reflectivity extraction with offset applied."""
        nradar = [sample_radar_dataset]
        gr_offset = 5.0

        refl, pir = get_gr_reflectivity(
            nradar, "corrected_reflectivity", gr_offset=gr_offset, gr_refl_threshold=10
        )

        # Check that non-masked values are reduced by offset (accounting for threshold masking)
        original = sample_radar_dataset["corrected_reflectivity"].values
        valid_mask = ~refl[0].mask
        assert np.all(refl[0].data[valid_mask] <= original[valid_mask])

    def test_has_valid_data_valid(self):
        """Test has_valid_data with valid array."""
        arr = np.ma.masked_array([10.0, 20.0, 30.0, 40.0, 50.0], mask=[False, False, False, False, False])
        assert has_valid_data(arr, min_valid=5) is True

    def test_has_valid_data_insufficient(self):
        """Test has_valid_data with insufficient samples."""
        arr = np.ma.masked_array([10.0, 20.0, 30.0], mask=[False, False, False])
        assert has_valid_data(arr, min_valid=5) is False

    def test_has_valid_data_all_nan(self):
        """Test has_valid_data with all NaN values."""
        arr = np.ma.masked_array([np.nan, np.nan, np.nan, np.nan, np.nan],
                                 mask=[True, True, True, True, True])
        assert has_valid_data(arr, min_valid=5) is False


# Integration tests
class TestVolumeMatchingIntegration:

    @pytest.fixture
    def test_data_dir(self):
        """Return path to test data directory."""
        # Adjust this path to where your test files are located
        return Path("tests/data")

    @pytest.fixture
    def radar_file(self, test_data_dir):
        """Path to test radar file."""
        return test_data_dir / "02_20181120_053630.pvol.h5"

    @pytest.fixture
    def gpm_file(self, test_data_dir):
        """Path to test GPM file."""
        return test_data_dir / "2A-CS-AUS-East.GPM.Ku.V8-20180723.20181120-S053341-E054123.026860.V06A.HDF5"

    @pytest.mark.skipif(not Path("tests/data").exists(), reason="Test data not available")
    def test_volume_matching_real_data(self, radar_file, gpm_file, temp_output_dir):
        """Test volume matching with real data files."""
        if not radar_file.exists() or not gpm_file.exists():
            pytest.skip("Test data files not found")

        # Use DBZH instead of corrected_reflectivity based on actual file structure
        result = volume_matching(
            gpmfile=str(gpm_file),
            grfile=str(radar_file),
            gr_offset=0,
            gr_beamwidth=1.0,
            radar_band="C",
            refl_name="DBZH",
            fname_prefix="test_radar"
        )

        assert isinstance(result, xr.Dataset)
        assert "refl_gpm_raw" in result
        assert "refl_gr_raw" in result
        assert "offset_found" in result.attrs
        assert "final_offset" in result.attrs
        assert result.attrs["gpm_orbit"] == 26860

    @pytest.mark.skipif(not Path("tests/data").exists(), reason="Test data not available")
    def test_vmatch_multi_pass_real_data(self, radar_file, gpm_file, temp_output_dir):
        """Test multi-pass volume matching with real data."""
        if not radar_file.exists() or not gpm_file.exists():
            pytest.skip("Test data files not found")

        # Use DBZH instead of corrected_reflectivity based on actual file structure
        vmatch_multi_pass(
            gpmfile=str(gpm_file),
            grfile=str(radar_file),
            gr_offset=0,
            gr_beamwidth=1.0,
            gr_refl_threshold=10,
            radar_band="C",
            refl_name="DBZH",
            fname_prefix="test_radar",
            offset_thld=0.5,
            output_dir=str(temp_output_dir)
        )

        # Check that output files were created
        first_pass_dir = temp_output_dir / "first_pass"
        final_pass_dir = temp_output_dir / "final_pass"

        assert first_pass_dir.exists()
        assert final_pass_dir.exists()

        first_pass_files = list(first_pass_dir.glob("*.nc"))
        final_pass_files = list(final_pass_dir.glob("*.nc"))

        assert len(first_pass_files) > 0
        assert len(final_pass_files) > 0

        # Load and check final result
        final_file = final_pass_files[0]
        final_dataset = xr.open_dataset(final_file)

        assert "iteration_number" in final_dataset.attrs
        assert "offset_history" in final_dataset.attrs

        final_dataset.close()


if __name__ == "__main__":
    pytest.main([__file__, "-v"])


"""
Various utilities for correction and conversion of satellite data.

@title: correct
@author: Valentin Louf <valentin.louf@bom.gov.au>
@institutions: Monash University and the Australian Bureau of Meteorology
@creation: 17/02/2020
@date: 12/05/2025

.. autosummary::
    :toctree: generated/

    compute_gaussian_curvature
    convert_gpmrefl_grband_dfr
    convert_gpmrefl_grband_phase_aware
    correct_parallax
    attenuation_correction_zphi
    attenuation_correction_gunn_east
    get_offset
    grid_displacement
"""

from typing import Tuple
import numpy as np
import xarray as xr


def compute_gaussian_curvature(lat0: float) -> float:
    """
    Determine the Earth's Gaussian radius of curvature at the radar
    https://en.wikipedia.org/wiki/Earth_radius#Radii_of_curvature

    Parameter:
    ----------
    lat0: float
        Ground radar latitude.

    Returns:
    --------
    ae: float
        Scaled Gaussian radius.
    """
    # Major and minor radii of the Ellipsoid
    a = 6378137.0  # Earth radius in meters
    e2 = 0.0066943800
    b = a * np.sqrt(1 - e2)

    tmp = (a * np.cos(np.pi / 180 * lat0)) ** 2 + (b * np.sin(np.pi / 180 * lat0)) ** 2  # Denominator
    an = (a**2) / np.sqrt(tmp)  # Radius of curvature in the prime vertical (east–west direction)
    am = (a * b) ** 2 / tmp**1.5  # Radius of curvature in the north–south meridian
    ag = np.sqrt(an * am)  # Earth's Gaussian radius of curvature
    ae = (4 / 3.0) * ag

    return ae


def convert_gpmrefl_grband_dfr(refl_gpm: np.ndarray, radar_band: str) -> np.ndarray:
    """
    Convert GPM reflectivity to ground radar band using the DFR relationship
    found in Louf et al. (2019) paper.

    Parameters:
    ===========
    refl_gpm:
        Satellite reflectivity field.
    radar_band: str
        Possible values are 'S', 'C', or 'X'

    Return:
    =======
    refl:
        Reflectivity conversion from Ku-band to ground radar band
    """
    if radar_band == "S":
        cof = np.array([2.01236803e-07, -6.50694273e-06, 1.10885533e-03, -6.47985914e-02, -7.46518423e-02])
        dfr = np.poly1d(cof)
    elif radar_band == "C":
        cof = np.array([1.21547932e-06, -1.23266138e-04, 6.38562875e-03, -1.52248868e-01, 5.33556919e-01])
        dfr = np.poly1d(cof)
    elif radar_band == "X":
        # Use of C band DFR relationship multiply by ratio
        cof = np.array([1.21547932e-06, -1.23266138e-04, 6.38562875e-03, -1.52248868e-01, 5.33556919e-01])
        dfr = 3.2 / 5.5 * np.poly1d(cof)
    else:
        raise ValueError(f"Radar reflectivity band ({radar_band}) not supported.")

    return refl_gpm + dfr(refl_gpm)


def convert_gpmrefl_grband_phase_aware(
    refl_gpm: np.ndarray,
    height_gpm: np.ndarray,
    height_bb: np.ndarray,
    radar_band: str,
    bb_margin: float = 500.0,
) -> np.ndarray:
    """
    Convert GPM Ku-band reflectivity to ground radar band using phase-aware DFR.

    This function accounts for hydrometeor phase (ice, melting layer, or liquid)
    by using the GPM bright band (melting layer) height. Different DFR relationships
    are applied based on whether the radar gate is above, within, or below the
    melting layer.

    Parameters:
    ===========
    refl_gpm: np.ndarray
        GPM Ku-band reflectivity field (dBZ). Shape: (nscan, nray, nbin)
    height_gpm: np.ndarray
        Height of each GPM gate (meters). Shape: (nscan, nray, nbin)
    height_bb: np.ndarray
        Bright band (melting layer) height from GPM (meters). Shape: (nscan, nray)
        Values < 0 indicate no bright band detected (all rain or all ice)
    radar_band: str
        Ground radar frequency band. Possible values: 'S', 'C', or 'X'
    bb_margin: float, optional
        Vertical margin around bright band height to define melting layer (meters).
        Default is 500m (±500m around heightBB defines the melting layer).

    Returns:
    ========
    refl_grband: np.ndarray
        Reflectivity converted from Ku-band to ground radar band (dBZ).
        Same shape as refl_gpm.

    References:
    ===========
    - Rain DFR coefficients from Louf et al. (2019) J. Atmos. Oceanic Technol.
    - Ice scattering theory: Tyynelä et al. (2011), Leinonen (2014)
    - GPM bright band detection: Awaka et al. (2016)
    """
    if radar_band not in ["S", "C", "X"]:
        raise ValueError(f"Radar reflectivity band ({radar_band}) not supported. Use 'S', 'C', or 'X'.")

    # Define rain DFR polynomial coefficients (from Louf et al. 2019)
    if radar_band == "S":
        rain_cof = np.array([2.01236803e-07, -6.50694273e-06, 1.10885533e-03, -6.47985914e-02, -7.46518423e-02])
    elif radar_band == "C":
        rain_cof = np.array([1.21547932e-06, -1.23266138e-04, 6.38562875e-03, -1.52248868e-01, 5.33556919e-01])
    elif radar_band == "X":
        rain_cof = np.array([1.21547932e-06, -1.23266138e-04, 6.38562875e-03, -1.52248868e-01, 5.33556919e-01])
        rain_cof = rain_cof * (3.2 / 5.5)

    rain_dfr = np.poly1d(rain_cof)

    # Ice DFR - smaller frequency dependence for ice particles
    # Based on Mie scattering theory for ice spheres
    # Ice has less frequency-dependent scattering than rain
    ice_dfr_offset = {
        'S': 0.3,   # dB - small positive offset for S-band
        'C': 0.5,   # dB - moderate offset for C-band
        'X': 0.8    # dB - larger offset for X-band
    }

    refl_grband = np.full_like(refl_gpm, np.nan, dtype=np.float64)
    # Expand heightBB to match 3D shape (nscan, nray, nbin)
    height_bb_3d = np.broadcast_to(height_bb[..., np.newaxis], refl_gpm.shape)

    no_bb_mask = height_bb_3d < 0
    liquid_mask = (height_bb_3d >= 0) & (height_gpm < (height_bb_3d - bb_margin))
    ice_mask = (height_bb_3d >= 0) & (height_gpm > (height_bb_3d + bb_margin))
    melting_mask = (height_bb_3d >= 0) & (height_gpm >= (height_bb_3d - bb_margin)) & (height_gpm <= (height_bb_3d + bb_margin))

    if np.any(no_bb_mask):
        refl_grband[no_bb_mask] = refl_gpm[no_bb_mask] + rain_dfr(refl_gpm[no_bb_mask])

    if np.any(liquid_mask):
        refl_grband[liquid_mask] = refl_gpm[liquid_mask] + rain_dfr(refl_gpm[liquid_mask])

    if np.any(ice_mask):
        refl_grband[ice_mask] = refl_gpm[ice_mask] + ice_dfr_offset[radar_band]

    if np.any(melting_mask):
        # Calculate normalized height within melting layer (0 = bottom, 1 = top)
        height_in_ml = height_gpm[melting_mask]
        bb_height = height_bb_3d[melting_mask]

        # Fraction of liquid (1 at bottom, 0 at top of melting layer)
        liquid_fraction = 1.0 - (height_in_ml - (bb_height - bb_margin)) / (2 * bb_margin)
        liquid_fraction = np.clip(liquid_fraction, 0, 1)

        # Interpolate between ice and rain DFR
        dfr_ice = ice_dfr_offset[radar_band]
        dfr_rain = rain_dfr(refl_gpm[melting_mask])
        dfr_mixed = liquid_fraction * dfr_rain + (1 - liquid_fraction) * dfr_ice

        refl_grband[melting_mask] = refl_gpm[melting_mask] + dfr_mixed

    return refl_grband


def attenuation_correction_zphi(zh: np.ndarray, kdp: np.ndarray, dr: float = 1, wavelength: str = "C") -> np.ndarray:
    """
    Attenuation correction using the ZPHI method.

    Parameters:
    -----------
    zh : array-like
        Horizontal reflectivity (dBZ)
    kdp : array-like
        Specific differential phase (deg/km)
    dr : float
        Range gate spacing in km
    wavelength : str
        'C' for C-band (~5.3 cm) or 'X' for X-band (~3.2 cm)

    Returns:
    --------
    zh_corrected : array
        Attenuation-corrected horizontal reflectivity (dBZ)
    """
    if dr > 10:
        raise ValueError("Range gate spacing 'dr' seems too large (>10 km). Check input.")

    if wavelength.upper() == "C":
        # C-band coefficients (Bringi et al. 2001)
        alpha_h = 0.08  # coefficient for Ah = alpha * Kdp^beta
        beta_h = 0.93

    elif wavelength.upper() == "X":
        # X-band coefficients (Park et al. 2005)
        alpha_h = 0.28
        beta_h = 0.95
    else:
        raise ValueError("Wavelength must be 'C' or 'X'")

    # Calculate specific attenuation
    ah = alpha_h * np.power(np.abs(kdp), beta_h) * np.sign(kdp)
    PIA_h = np.nancumsum(ah * dr, axis=1)  # Path Integrated Attenuation for Zh
    zh_corrected = zh + 2 * PIA_h  # Factor of 2 for two-way path

    return zh_corrected


def attenuation_correction_gunn_east(
    zh: np.ndarray, dr: float = 1, wavelength: str = "C", max_thld: float = 10
) -> np.ndarray:
    """
    Attenuation correction using Gunn and East (1954) method.
    This is based on the power-law relationship.
    Eq. 2.66 p 106 of radar meteorology by Henri Sauvagot.

    Parameters:
    -----------
    zh : array-like
        Horizontal reflectivity in dBZ along a radial
    dr : float
        Range gate spacing in km
    wavelength : str
        'C' for C-band or 'X' for X-band
    max_thld: float
        Capped-attenuation correction to avoid blowing up the refl in dB.

    Returns:
    --------
    zh_corrected : array
        Attenuation-corrected reflectivity (dBZ)
    """
    if dr > 10:
        raise ValueError("Range gate spacing 'dr' seems too large (>10 km). Check input.")
    if wavelength.upper() == "C":
        l = 5.5
    elif wavelength.upper() == "X":
        l = 3.2
    else:
        raise ValueError("Wavelength must be 'C' or 'X'")

    # Convert dBZ to linear Z (mm^6/m^3)
    Z_linear = 10 ** (zh / 10.0)
    R = (Z_linear / 200) ** (1 / 1.6)
    ap = 0.35e-2 * R**1.6 / (l**4) + 0.22e-2 * R / l

    PIA_h = np.nancumsum(ap * dr, axis=1)
    PIA_h[PIA_h > max_thld / 2] = max_thld / 2
    Z_corrected = zh + 2 * PIA_h

    return Z_corrected


def correct_parallax(
    sr_x: np.ndarray, sr_y: np.ndarray, gpmset: xr.Dataset
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Adjust the geo-locations of the SR pixels. The `sr_xy` coordinates of the
    SR beam footprints need to be in the azimuthal equidistant projection of
    the ground radar. This ensures that the ground radar is fixed at
    xy-coordinate (0, 0), and every SR bin has its relative xy-coordinates
    with respect to the ground radar site.

    Parameters:
    -----------
    sr_x: ndarray
        Array of x-coordinates of shape (nscans, nbeams)
    sr_y: ndarray
        Array of y-coordinates of shape (nscans, nbeams)
    gpmset: xarray

    Returns:
    --------
    sr_xyp : :class:`numpy:numpy.ndarray`
        Array of parallax corrected coordinates
        of shape (nscans, nbeams, nbins).
    z_sr : :class:`numpy:numpy.ndarray`
        Array of SR bin altitudes of shape (nscans, nbeams, nbins).
    """
    r_sr_inv, alpha = gpmset.nbin.values, gpmset.nray.values
    # calculate height of bin
    z = r_sr_inv * np.cos(np.deg2rad(alpha))[..., np.newaxis]
    z_sr = np.repeat(z[np.newaxis, :, :], len(gpmset.nscan), axis=0)
    # calculate bin ground xy-displacement length
    ds = r_sr_inv * np.sin(np.deg2rad(alpha))[..., np.newaxis]

    # calculate x,y-differences between ground coordinate
    # and center ground coordinate [25th element]
    center = int(np.floor(len(sr_x[-1]) / 2.0))
    xdiff = sr_x - sr_x[:, center][:, np.newaxis]
    ydiff = sr_y - sr_y[:, center][:, np.newaxis]

    # assuming ydiff and xdiff being a triangles adjacent and
    # opposite this calculates the xy-angle of the SR scan
    ang = np.arctan2(ydiff, xdiff)

    # calculate displacement dx, dy from displacement length
    dx = ds * np.cos(ang)[..., np.newaxis]
    dy = ds * np.sin(ang)[..., np.newaxis]

    # subtract displacement from SR ground coordinates
    sr_xp = sr_x[..., np.newaxis] - dx
    sr_yp = sr_y[..., np.newaxis] - dy

    return sr_xp, sr_yp, z_sr


def get_offset(matchset: xr.Dataset, dr: float, nbins: int = 200) -> float:
    """
    Compute the Offset between GR and GPM. It will try to compute the mode of
    the distribution and if it fails, then it will use the mean.

    Parameter:
    ==========
    matchset: xr.Dataset
        Dataset of volume matching.
    dr: float
        Ground radar gate spacing (m).
    nbins: int
        Defines the number of equal-width bins in the distribution.

    Returns:
    ========
    offset: float
        Offset between GR and GPM
    """
    offset = np.arange(-15, 15, 0.2, dtype=np.float64)
    area = np.zeros_like(offset)

    refl_gpm = matchset.refl_gpm_grband.values.flatten().copy()
    refl_gr = matchset.refl_gr_weigthed.values.flatten().copy()
    fmin = matchset.fmin_gr.values.flatten().copy()

    pos = (refl_gpm > 36) | (refl_gr > 36) | (fmin != 1)
    if (~pos).sum() == 0:  # Relaxing fmin parameter.
        pos = (refl_gpm > 36) | (refl_gr > 36) | (fmin < 0.9)
    if (~pos).sum() == 0:
        pos = (refl_gpm > 36) | (refl_gr > 36) | (fmin < 0.7)

    refl_gpm[pos] = np.nan
    refl_gr[pos] = np.nan

    pdf_gpm, _ = np.histogram(refl_gpm, range=(0, 50), bins=nbins, density=True)
    for idx, a in enumerate(offset):
        pdf_gr, _ = np.histogram(refl_gr - a, range=(0, 50), bins=nbins, density=True)
        diff = np.min([pdf_gr, pdf_gpm], axis=0)
        area[idx] = np.sum(diff)

    smoothed_area = np.convolve([1] * 12, area, "same")
    maxpos = np.argmax(smoothed_area)
    gr_offset = offset[maxpos]
    return gr_offset


def grid_displacement(field1: np.ndarray, field2: np.ndarray) -> Tuple[Tuple[int, int], Tuple[int, int]]:
    """
    Calculate the grid displacement using Phase correlation.
    http://en.wikipedia.org/wiki/Phase_correlation

    Parameters:
    -----------
    field1, field2 : ndarray
       Fields separated in time.

    Returns:
    --------
    displacement : two-tuple
         integers if pixels, otherwise floats. Result of the calculation
    """
    # create copies of the data
    ige1 = np.ma.masked_invalid(10 ** (field1 / 10)).filled(0)
    ige2 = np.ma.masked_invalid(10 ** (field2 / 10)).filled(0)

    # discrete fast fourier transformation and complex conjugation of image 2
    image1FFT = np.fft.fft2(ige1)
    image2FFT = np.conjugate(np.fft.fft2(ige2))

    # inverse fourier transformation of product -> equal to cross correlation
    imageCCor = np.real(np.fft.ifft2((image1FFT * image2FFT)))

    # Shift the zero-frequency component to the center of the spectrum
    imageCCorShift = np.fft.fftshift(imageCCor)
    row, col = ige1.shape

    # find the peak in the correlation
    yShift, xShift = np.unravel_index(np.argmax(imageCCorShift), (row, col))
    yShift -= row // 2
    xShift -= col // 2

    return (xShift, yShift)



"""
Volume matching of ground radar and GPM satellite. It also works with the
latest version of TRMM data.

@title: gpmmatch
@author: Valentin Louf <valentin.louf@bom.gov.au>
@institutions: Monash University and the Australian Bureau of Meteorology
@creation: 17/02/2020
@date: 12/12/2025

.. autosummary::
    :toctree: generated/

    NoRainError
    get_radar_coordinates
    get_gr_reflectivity
    volume_matching
    vmatch_multi_pass
"""

import uuid
import datetime
import warnings
import itertools
from pathlib import Path
from typing import List, Tuple, Union

import numpy as np
import pandas as pd
import xarray as xr
from scipy.spatial import cKDTree

from .correct import get_offset
from .io import data_load_and_checks
from .default import get_metadata

# Constants
MIN_SAMPLE_POINTS = 20
MIN_REFL_SAMPLES = 5
MAX_OFFSET_THRESHOLD = 15


class NoRainError(Exception):
    pass


def generate_filename(radar: xr.Dataset, gpmset: xr.Dataset, fname_prefix: str) -> str:
    """
    Generates a filename for the output dataset based on the input GPM and ground radar files.

    Parameters:
    ----------
    radar: xr.Dataset
        Ground radar dataset.
    gpmset: xr.Dataset
        GPM dataset.
    fname_prefix: str
        Name of the ground radar to use as label for the output file.

    Returns:
    -------
    str
        Generated filename for the output dataset.
    """
    date = pd.Timestamp(radar.time[0].values).strftime("%Y%m%d.%H%M")
    outfilename = f"vmatch.gpm.orbit.{gpmset.attrs['orbit']:07}.{fname_prefix}.{date}.nc"
    return outfilename


def get_radar_coordinates(
    nradar: List[xr.Dataset], elevation_offset: Union[float, None] = None
) -> Tuple[np.ndarray, np.ndarray, List[np.ndarray], List[np.ndarray], List[np.ndarray]]:
    """
    Extracts the ground radar coordinates and elevation angles.

    Parameters:
    ----------
    nradar: list of xarray.Dataset
        List of ground radar datasets for each tilt (pyodim structure).
    elevation_offset: float, optional
        Offset to add to the elevation angles of the ground radar data.

    Returns:
    -------
    range_gr: np.ndarray
        Array of ground radar range values.
    elev_gr: np.ndarray
        Array of ground radar elevation angles.
    xradar: List[np.ndarray]
        List of x coordinates of the ground radar for each tilt.
    yradar: List[np.ndarray]
        List of y coordinates of the ground radar for each tilt.
    time_radar: List[np.ndarray]
        List of time values of the ground radar for each tilt.
    """
    range_gr = nradar[0].range.values
    elev_gr = np.array([r.elevation.values[0] for r in nradar])

    if elevation_offset is not None:
        print(f"Correcting the GR elevation by an offset of {elevation_offset}.")
        elev_gr += elevation_offset

    xradar = [r.x.values for r in nradar]
    yradar = [r.y.values for r in nradar]
    time_radar = [r.time.values for r in nradar]

    return range_gr, elev_gr, xradar, yradar, time_radar


def get_gr_reflectivity(
    nradar: List[xr.Dataset], refl_name: str, gr_offset: float, gr_refl_threshold: float
) -> Tuple[List[np.ma.MaskedArray], List[np.ndarray]]:
    """
    Extracts the ground radar reflectivity and computes the path-integrated reflectivity.

    Parameters:
    ----------
    nradar: list of xarray.Dataset
        List of ground radar datasets.
    refl_name: str
        Name of the reflectivity field in the ground radar data.
    gr_offset: float
        Offset to add to the reflectivity of the ground radar data.
    gr_refl_threshold: float
        Minimum reflectivity threshold on ground radar data.

    Returns:
    -------
    ground_radar_reflectivity: List[np.ma.MaskedArray]
        Array of ground radar reflectivity values for each radar tilt.
    pir_gr: List[np.ndarray]
        Array of path-integrated reflectivity values for ground radar for each radar tilt.
    """
    ground_radar_reflectivity = []
    pir_gr = []

    using_corrected_field = False
    for radar in nradar:
        if "ZH_ATTEN_CORR" in radar.data_vars:
            refl = radar["ZH_ATTEN_CORR"].values.copy() - gr_offset
            using_corrected_field = True
        else:
            refl = radar[refl_name].values.copy() - gr_offset
        refl[refl < gr_refl_threshold] = np.nan
        refl = np.ma.masked_invalid(refl)
        ground_radar_reflectivity.append(refl)

        dr = radar.range[1].values - radar.range[0].values  # Range resolution in meters
        pir = 10 * np.log10(np.cumsum((10 ** (refl / 10)).filled(0), axis=1) * dr)
        pir_gr.append(pir)

    if using_corrected_field:
        print("Using attenuation-corrected reflectivity field for ground radar.")

    return ground_radar_reflectivity, pir_gr


def has_valid_data(arr: np.ma.MaskedArray, min_valid: int = MIN_REFL_SAMPLES) -> bool:
    """
    Check if array has sufficient valid (non-NaN) data.

    Parameters:
    ----------
    arr: np.ma.MaskedArray
        Array to check for valid data.
    min_valid: int
        Minimum number of valid samples required.

    Returns:
    -------
    bool
        True if array has sufficient valid data.
    """
    return len(arr) >= min_valid and not np.all(np.isnan(arr.filled(np.nan)))


def volume_matching(
    gpmfile: str,
    grfile: str,
    grfile2: Union[str, None] = None,
    gr_offset: float = 0,
    gr_beamwidth: float = 1,
    gr_rmax: Union[float, None] = None,
    gr_refl_threshold: float = 10,
    radar_band: str = "C",
    refl_name: str = "corrected_reflectivity",
    correct_attenuation: bool = True,
    elevation_offset: Union[float, None] = None,
    fname_prefix: Union[str, None] = None,
    kdp_name: Union[str, None] = "KDP",
    phase_aware_dfr: bool = True,
) -> xr.Dataset:
    """
    Performs the volume matching of GPM satellite data to ground based radar.

    Parameters:
    ----------
    gpmfile: str
        GPM data file.
    grfile: str
        Ground radar input file.
    grfile2: str, optional
        Second ground radar input file to compute the advection.
    gr_offset: float
        Offset to add to the reflectivity of the ground radar data.
    gr_beamwidth: float
        Ground radar 3dB-beamwidth.
    gr_rmax: float
        Ground radar maximum range in meters (100,000 m). Actual max range used (up to 250,000 m).
    gr_refl_threshold: float
        Minimum reflectivity threshold on ground radar data.
    radar_band: str
        Ground radar frequency band.
    refl_name: str
        Name of the reflectivity field in the ground radar data.
    correct_attenuation: bool
        Should we correct for C- or X-band ground radar attenuation
    elevation_offset: float
        Adding an offset in case the elevation angle needs to be corrected.
    fname_prefix: str
        Name of the ground radar to use as label for the output file.
    kdp_name: Optional[str]
        Name of the KDP field in the ground radar data for attenuation correction.
    phase_aware_dfr: bool
        Use phase-aware DFR conversion that accounts for ice, melting layer, and
        liquid precipitation phases using GPM bright band height. Default is True.

    Returns:
    --------
    matchset: xarray.Dataset
        Dataset containing the matched GPM and ground radar data.
    """
    if fname_prefix is None:
        fname_prefix = "unknown_radar"

    gpmset, nradar = data_load_and_checks(
        gpmfile,
        grfile,
        refl_name=refl_name,
        correct_attenuation=correct_attenuation,
        radar_band=radar_band,
        kdp_name=kdp_name,
        phase_aware_dfr=phase_aware_dfr,
    )

    nprof = gpmset.precip_in_gr_domain.values.sum()
    ntilt = len(nradar)

    ground_radar_reflectivity, pir_gr = get_gr_reflectivity(nradar, refl_name, gr_offset, gr_refl_threshold)
    range_gr, elev_gr, xradar, yradar, tradar = get_radar_coordinates(nradar, elevation_offset)
    dr = range_gr[1] - range_gr[0]
    if gr_rmax is None:
        gr_rmax = range_gr.max() if range_gr.max() < 250e3 else 250e3

    # Cache nradar attributes
    nradar_range = [nr.range.values for nr in nradar]
    nradar_azimuth = [nr.azimuth.values for nr in nradar]

    # Extract and cache GPM data
    position_precip_domain = gpmset.precip_in_gr_domain.values != 0
    gpm_nray = gpmset.nray.values
    gpm_nscan = gpmset.nscan.values
    gpm_elev_from_gr = gpmset.elev_from_gr.values
    gpm_x = gpmset.x.values
    gpm_y = gpmset.y.values
    gpm_z = gpmset.z.values
    gpm_dr = gpmset.dr
    gpm_beamwidth = gpmset.beamwidth
    gpm_altitude = gpmset.altitude
    gpm_earth_radius = gpmset.earth_gaussian_radius
    gpm_distance_from_sr = gpmset.distance_from_sr.values
    gpm_zfactor = gpmset.zFactorCorrected.values
    gpm_refl_grband = gpmset.reflectivity_grband.values
    gpm_overpass_time = gpmset.overpass_time.values

    alpha, _ = np.meshgrid(gpm_nray, gpm_nscan)
    alpha = alpha[position_precip_domain]

    elev_sat = gpm_elev_from_gr[position_precip_domain]
    xsat = gpm_x[position_precip_domain]
    ysat = gpm_y[position_precip_domain]
    zsat = gpm_z[position_precip_domain]

    rsat = np.zeros(gpm_zfactor.shape)
    for i in range(rsat.shape[0]):
        rsat[i, :] = gpm_distance_from_sr

    volsat = 1e-9 * gpm_dr * (rsat[position_precip_domain] * np.deg2rad(gpm_beamwidth)) ** 2

    refl_gpm_raw = np.ma.masked_invalid(gpm_zfactor[position_precip_domain])
    reflectivity_gpm_grband = np.ma.masked_invalid(gpm_refl_grband[position_precip_domain])

    # Compute Path-integrated reflectivities
    pir_gpm = 10 * np.log10(np.cumsum((10 ** (np.ma.masked_invalid(refl_gpm_raw) / 10)).filled(0), axis=-1) * dr)
    pir_gpm = np.ma.masked_invalid(pir_gpm)

    # Pre-compute and cache everything
    beamwidth_rad = np.deg2rad(gr_beamwidth)
    half_beamwidth = gr_beamwidth / 2

    # Pre-compute for each tilt
    R_list = []
    DT_list = []
    volgr_list = []
    kdtree_list = []
    xradar_flat_list = []
    yradar_flat_list = []
    gr_refl_flat_list = []
    pir_gr_flat_list = []
    orig_shapes = []

    for jj in range(ntilt):
        deltat = tradar[jj] - gpm_overpass_time
        R, _ = np.meshgrid(nradar_range[jj], nradar_azimuth[jj])
        _, DT = np.meshgrid(nradar_range[jj], deltat)
        volgr = 1e-9 * dr * (R * beamwidth_rad) ** 2

        R_list.append(R)
        DT_list.append(DT)
        volgr_list.append(volgr)

        # Flatten and cache all arrays
        x_flat = xradar[jj].ravel()
        y_flat = yradar[jj].ravel()
        xradar_flat_list.append(x_flat)
        yradar_flat_list.append(y_flat)

        orig_shapes.append(xradar[jj].shape)

        # Build KDTree
        xy_coords = np.column_stack([x_flat, y_flat])
        kdtree_list.append(cKDTree(xy_coords))

        # Pre-flatten data arrays
        gr_refl_flat_list.append(ground_radar_reflectivity[jj].ravel())
        pir_gr_flat_list.append(pir_gr[jj].ravel())

    # Initialize output
    datakeys = [
        "refl_gpm_raw",
        "refl_gr_weigthed",
        "refl_gpm_grband",
        "pir_gpm",
        "pir_gr",
        "refl_gr_raw",
        "std_refl_gpm",
        "std_refl_gr",
        "sample_gpm",
        "reject_gpm",
        "fmin_gpm",
        "fmin_gr",
        "sample_gr",
        "reject_gr",
        "volume_match_gpm",
        "volume_match_gr",
    ]
    data = {k: np.full((nprof, ntilt), np.nan) for k in datakeys}

    x = np.zeros((nprof, ntilt))
    y = np.zeros((nprof, ntilt))
    z = np.zeros((nprof, ntilt))
    r = np.zeros((nprof, ntilt))
    dz = np.zeros((nprof, ntilt))
    ds = np.zeros((nprof, ntilt))
    delta_t = np.full((nprof, ntilt), np.nan)

    # Optimized main loop
    for ii, jj in itertools.product(range(nprof), range(ntilt)):
        if elev_gr[jj] - half_beamwidth < 0:
            continue

        epos = (elev_sat[ii, :] >= elev_gr[jj] - half_beamwidth) & (elev_sat[ii, :] <= elev_gr[jj] + half_beamwidth)
        if not np.any(epos):
            continue

        x[ii, jj] = x_mean = np.mean(xsat[ii, epos])
        y[ii, jj] = y_mean = np.mean(ysat[ii, epos])
        z[ii, jj] = z_mean = np.mean(zsat[ii, epos])

        data["sample_gpm"][ii, jj] = n_epos = np.sum(epos)
        data["volume_match_gpm"][ii, jj] = np.sum(volsat[ii, epos])

        alpha_ii = alpha[ii]
        cos_alpha = np.cos(np.deg2rad(alpha_ii))

        dz[ii, jj] = n_epos * gpm_dr * cos_alpha
        ds[ii, jj] = np.deg2rad(gpm_beamwidth) * np.mean(gpm_altitude - zsat[ii, epos]) / cos_alpha

        # Calculate s_sat for this specific profile
        s_sat_ii = np.sqrt(x_mean**2 + y_mean**2)
        r[ii, jj] = (gpm_earth_radius + z_mean) * np.sin(s_sat_ii / gpm_earth_radius) / np.cos(np.deg2rad(elev_gr[jj]))

        if r[ii, jj] + ds[ii, jj] / 2 > gr_rmax:
            continue
        if np.isnan(x_mean) or np.isnan(y_mean) or np.isnan(ds[ii, jj]):
            continue

        # KDTree query with pre-allocated arrays
        search_radius = ds[ii, jj] / 2
        indices_flat = kdtree_list[jj].query_ball_point([x_mean, y_mean], search_radius)

        if not indices_flat:
            continue

        # Use flat indices directly - faster than unravel
        xradar_subset = xradar_flat_list[jj][indices_flat]
        yradar_subset = yradar_flat_list[jj][indices_flat]

        # Vectorized distance calculation
        dx = xradar_subset - x_mean
        dy = yradar_subset - y_mean
        roi_gr_at_vol = np.sqrt(dx * dx + dy * dy)

        # Get data using flat indices
        volgr_subset = volgr_list[jj].ravel()[indices_flat]
        R_subset = R_list[jj].ravel()[indices_flat]
        DT_subset = DT_list[jj].ravel()[indices_flat]

        # Weight calculation
        normalized_distance = roi_gr_at_vol / search_radius
        w = volgr_subset * np.exp(-(normalized_distance**2))

        # Extract reflectivity
        refl_gpm = refl_gpm_raw[ii, epos]
        refl_gpm_grband = reflectivity_gpm_grband[ii, epos]
        refl_gr_raw = gr_refl_flat_list[jj][indices_flat]
        pir_gr_flat = pir_gr_flat_list[jj][indices_flat]

        try:
            delta_t[ii, jj] = np.max(DT_subset)
        except ValueError:
            continue

        if not (has_valid_data(refl_gpm) and has_valid_data(refl_gr_raw)):
            continue

        # Statistics - use compressed() for masked arrays to avoid overhead
        refl_gpm_compressed = refl_gpm.compressed()
        refl_gr_compressed = refl_gr_raw.compressed()

        data["fmin_gpm"][ii, jj] = (
            np.sum(refl_gpm_compressed > 0) / len(refl_gpm_compressed) if len(refl_gpm_compressed) > 0 else np.nan
        )
        data["fmin_gr"][ii, jj] = np.sum(~refl_gr_raw.mask & (refl_gr_raw >= gr_refl_threshold)) / len(refl_gr_raw)

        data["refl_gpm_raw"][ii, jj] = np.mean(refl_gpm_compressed) if len(refl_gpm_compressed) > 0 else np.nan
        data["refl_gpm_grband"][ii, jj] = np.ma.mean(refl_gpm_grband)
        data["pir_gpm"][ii, jj] = np.ma.mean(pir_gpm[ii, epos])
        data["std_refl_gpm"][ii, jj] = np.std(refl_gpm_compressed) if len(refl_gpm_compressed) > 0 else np.nan
        data["reject_gpm"][ii, jj] = n_epos - len(refl_gpm_compressed)

        data["volume_match_gr"][ii, jj] = np.sum(volgr_subset)

        # Weighted mean - avoid creating new masked arrays
        valid_mask = ~refl_gr_raw.mask if np.ma.is_masked(refl_gr_raw) else np.ones(len(refl_gr_raw), dtype=bool)
        if np.any(valid_mask):
            data["refl_gr_weigthed"][ii, jj] = np.sum(w[valid_mask] * refl_gr_raw.data[valid_mask]) / np.sum(
                w[valid_mask]
            )

        data["refl_gr_raw"][ii, jj] = np.mean(refl_gr_compressed) if len(refl_gr_compressed) > 0 else np.nan
        data["pir_gr"][ii, jj] = np.ma.mean(pir_gr_flat)
        data["std_refl_gr"][ii, jj] = np.std(refl_gr_compressed) if len(refl_gr_compressed) > 0 else np.nan
        data["reject_gr"][ii, jj] = len(indices_flat)
        data["sample_gr"][ii, jj] = len(refl_gr_compressed)

    data["x"] = x
    data["y"] = y
    data["z"] = z
    data["r"] = r
    data["nprof"] = np.arange(nprof, dtype=np.int32)
    data["ntilt"] = np.arange(ntilt, dtype=np.int32)
    data["elevation_gr"] = elev_gr[:ntilt]
    data["timedelta"] = delta_t

    if np.sum((~np.isnan(data["refl_gpm_raw"])) & (~np.isnan(data["refl_gr_raw"]))) < MIN_SAMPLE_POINTS:
        raise NoRainError(f"At least {MIN_SAMPLE_POINTS} sample points are required.")

    # Transform to xarray
    match = {}
    for k, v in data.items():
        if k in ["ntilt", "elevation_gr"]:
            match[k] = (("ntilt"), v)
        elif k == "nprof":
            match[k] = (("nprof"), v)
        else:
            match[k] = (("nprof", "ntilt"), np.ma.masked_invalid(v.astype(np.float32)))

    matchset = xr.Dataset(match)
    metadata = get_metadata()
    for k, v in metadata.items():
        for sk, sv in v.items():
            try:
                matchset[k].attrs[sk] = sv
            except KeyError:
                continue

    ar = gpm_x**2 + gpm_y**2
    iscan, _, _ = np.where(ar == ar.min())
    gpm_overpass_time_iso = pd.Timestamp(gpm_nscan[iscan[0]]).isoformat()
    gpm_mindistance = np.sqrt(gpm_x**2 + gpm_y**2)[:, :, 0][gpmset.flagPrecip.values > 0].min()
    offset = get_offset(matchset, dr)
    if np.abs(offset) > MAX_OFFSET_THRESHOLD:
        raise ValueError(f"Offset of {offset} dB for {grfile} too big to mean anything.")

    matchset.attrs["offset_applied"] = gr_offset
    matchset.attrs["offset_found"] = offset
    matchset.attrs["final_offset"] = gr_offset + offset
    matchset.attrs["estimated_calibration_offset"] = f"{offset:0.4} dB"
    matchset.attrs["gpm_overpass_time"] = gpm_overpass_time_iso
    matchset.attrs["gpm_min_distance"] = np.round(gpm_mindistance)
    matchset.attrs["gpm_orbit"] = gpmset.attrs["orbit"]
    matchset.attrs["radar_start_time"] = nradar[0].attrs["start_time"]
    matchset.attrs["radar_end_time"] = nradar[0].attrs["end_time"]
    matchset.attrs["radar_longitude"] = nradar[0].attrs["longitude"]
    matchset.attrs["radar_latitude"] = nradar[0].attrs["latitude"]
    matchset.attrs["radar_range_res"] = dr
    matchset.attrs["radar_beamwidth"] = gr_beamwidth
    matchset.attrs["country"] = "Australia"
    matchset.attrs["creator_email"] = "valentin.louf@bom.gov.au"
    matchset.attrs["creator_name"] = "Valentin Louf"
    matchset.attrs["date_created"] = datetime.datetime.now().isoformat()
    matchset.attrs["uuid"] = str(uuid.uuid4())
    matchset.attrs["institution"] = "Bureau of Meteorology"
    matchset.attrs["references"] = "doi:10.1175/JTECH-D-18-0007.1 ; doi:10.1175/JTECH-D-17-0128.1"
    matchset.attrs["disclaimer"] = (
        "If you are using this data/technique for a scientific publication, please cite the papers given in references."
    )
    matchset.attrs["naming_authority"] = "au.org.nci"
    matchset.attrs["summary"] = "GPM volume matching technique."
    matchset.attrs["filename"] = generate_filename(nradar[0], gpmset, fname_prefix)

    return matchset


def vmatch_multi_pass(
    gpmfile: str,
    grfile: str,
    grfile2: Union[str, None] = None,
    gr_offset: float = 0,
    gr_beamwidth: float = 1,
    gr_rmax: Union[float, None] = None,
    gr_refl_threshold: float = 10,
    radar_band: str = "C",
    refl_name: str = "corrected_reflectivity",
    correct_attenuation: bool = True,
    elevation_offset: Union[float, None] = None,
    fname_prefix: Union[str, None] = None,
    offset_thld: float = 0.5,
    output_dir: Union[str, None] = None,
    kdp_name: Union[str, None] = "KDP",
    phase_aware_dfr: bool = True,
) -> None:
    """
    Multi-pass volume matching driver function with offset computation.

    Parameters:
    ----------
    gpmfile: str
        GPM data file.
    grfile: str
        Ground radar input file.
    grfile2: str
        Second ground radar input file to compute the advection.
    gr_offset: float
        Offset to add to the reflectivity of the ground radar data.
    gr_beamwidth: float
        Ground radar 3dB-beamwidth.
    gr_rmax: float
        Ground radar maximum range in meters (100,000 m).
    gr_refl_threshold: float
        Minimum reflectivity threshold on ground radar data.
    radar_band: str
        Ground radar frequency band.
    refl_name: str
        Name of the reflectivity field in the ground radar data.
    fname_prefix: str
        Name of the ground radar to use as label for the output file.
    correct_attenuation: bool
        Should we correct for C- or X-band ground radar attenuation
    elevation_offset: float
        Adding an offset in case the elevation angle needs to be corrected.
    offset_thld: float
        Offset threshold (in dB) between GPM and GR to stop the iteration.
    output_dir: str
        Path to output directory.
    kdp_name: Optional[str]
        Name of the KDP field in the ground radar data for attenuation correction.
    phase_aware_dfr: bool
        Use phase-aware DFR conversion that accounts for ice, melting layer, and
        liquid precipitation phases using GPM bright band height. Default is True.
    """

    def _save(dset: xr.Dataset, output_directory: str) -> None:
        """
        Generate multipass metadata and file name.
        """
        dset.attrs["iteration_number"] = counter
        dset.attrs["offset_history"] = ",".join([f"{float(i):0.3}" for i in offset_keeping_track])
        filename = dset.attrs["filename"].replace(".nc", f".pass{counter}.nc")

        outfilename = Path(output_directory) / filename
        print(f"Saving {outfilename.name} to {output_directory}.")
        dset.to_netcdf(str(outfilename), encoding={k: {"zlib": True} for k in dset.data_vars})

        # Check if the file was created successfully.
        if outfilename.exists():
            print(f"{outfilename.name} written to {output_directory}.")
        else:
            print(f"Error writing {outfilename.name} to {output_directory}.")
            raise FileNotFoundError(f"File {outfilename} could not be created.")

        return None

    counter = 0
    if fname_prefix is None:
        fname_prefix = "unknown_radar"
        print(f"No 'fname_prefix' defined. The output files will be named {fname_prefix}")
    if output_dir is None:
        output_dir = str(Path.cwd())
        print(f"No 'output_dir' defined. The output files will be saved {output_dir}")
    if correct_attenuation and radar_band not in ["C", "X"]:
        print(
            f"Attenuation correction is only available for C- and X-band radars. Setting 'correct_attenuation' to False."
        )
        correct_attenuation = False

    # Generate output directories.
    output_dirs = {
        "first": Path(output_dir) / "first_pass",
        "final": Path(output_dir) / "final_pass",
    }
    for dir_path in output_dirs.values():
        dir_path.mkdir(parents=True, exist_ok=True)

    # Function arguments dictionary.
    kwargs = {
        "gpmfile": gpmfile,
        "grfile": grfile,
        "grfile2": grfile2,
        "gr_offset": gr_offset,
        "radar_band": radar_band,
        "refl_name": refl_name,
        "fname_prefix": fname_prefix,
        "correct_attenuation": correct_attenuation,
        "gr_beamwidth": gr_beamwidth,
        "gr_rmax": gr_rmax,
        "gr_refl_threshold": gr_refl_threshold,
        "elevation_offset": elevation_offset,
        "kdp_name": kdp_name,
        "phase_aware_dfr": phase_aware_dfr,
    }

    # First pass
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        matchset = volume_matching(**kwargs)
    pass_offset = matchset.attrs["offset_found"]
    kwargs["gr_offset"] = pass_offset  # Update offset in kwargs for next pass
    offset_keeping_track = [pass_offset]
    final_offset_keeping_track = [matchset.attrs["final_offset"]]
    _save(matchset, str(output_dirs["first"]))

    if np.isnan(pass_offset):
        dtime = matchset.attrs["gpm_overpass_time"]
        print(f"Offset is NAN for pass {counter} on {dtime}.")
        return None

    # Multiple pass as long as the difference is more than threshold or counter is 6
    if np.abs(pass_offset) > offset_thld:
        for counter in range(1, 6):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                new_matchset = volume_matching(**kwargs)

            # Check offset found.
            gr_offset = new_matchset.attrs["final_offset"]
            kwargs["gr_offset"] = gr_offset  # Update offset in kwargs for next pass
            pass_offset = new_matchset.attrs["offset_found"]

            if np.isnan(pass_offset):
                # Solution converged already. Using previous iteration as final result.
                counter -= 1
                break
            if (np.abs(pass_offset) > np.abs(offset_keeping_track[-1])) and (counter > 1):
                counter -= 1
                break

            # Pass results are good enough to continue.
            matchset = new_matchset
            offset_keeping_track.append(pass_offset)
            final_offset_keeping_track.append(gr_offset)
            if np.abs(pass_offset) < offset_thld:
                break

    # Save final iteration.
    _save(matchset, str(output_dirs["final"]))
    return None
